import { MOS6502 } from "../common/cpu/MOS6502";
import { BasicMachine, RasterFrameBased } from "../common/devices";
declare class CIA {
    regs: Uint8Array;
    reset(): void;
    read(a: number): number;
    write(a: number, v: number): void;
}
declare class VIC_II {
    platform: C64;
    scanline: number;
    regs: Uint8Array;
    cram: Uint8Array;
    pixels: Uint8Array;
    cycle: number;
    vc: number;
    vcbase: number;
    rc: number;
    vmli: number;
    constructor(platform: any);
    reset(): void;
    read(a: number): number;
    write(a: number, v: number): void;
    setScanline(sl: number): void;
    saveState(): {
        regs: Uint8Array;
        cram: Uint8Array;
    };
    loadState(s: any): void;
    c_access(): number;
    g_access(data: number): number;
    getx(): number;
    clockPulse(): void;
    doDMA(): number;
    doInterrupt(): boolean;
    static stateToLongString(state: any): string;
}
export declare class C64 extends BasicMachine implements RasterFrameBased {
    cpuFrequency: number;
    canvasWidth: number;
    overscan: boolean;
    numTotalScanlines: number;
    numVisibleScanlines: number;
    defaultROMSize: number;
    cpuCyclesPerLine: number;
    cpuCyclesPreDMA: number;
    sampleRate: number;
    cpu: MOS6502;
    ram: Uint8Array;
    rom: Uint8Array;
    bios: Uint8Array;
    vic: VIC_II;
    cia1: CIA;
    cia2: CIA;
    probeDMABus: any;
    lastFrameCycles: number;
    enableKERNAL: boolean;
    enableIO: boolean;
    enableBASIC: boolean;
    enableCART: boolean;
    constructor();
    read: (a: number, v?: number) => number;
    write: (a: number, v?: number) => number;
    getKeyboardMap(): any;
    getKeyboardFunction(): (o: any, key: any, code: any, flags: any) => void;
    write6510(a: number, v: number): void;
    writeIO(a: number, v: number): void;
    readIO(a: number): number;
    readVRAMAddress(a: number): number;
    advanceFrame(trap: any): number;
    getRasterX(): number;
    getRasterY(): number;
    loadROM(data: any): void;
    loadBIOS(data: any): void;
    reset(): void;
    readConst(addr: number): number;
    loadState(state: any): void;
    saveState(): {
        c: any;
        ram: Uint8Array;
        vic: {
            regs: Uint8Array;
            cram: Uint8Array;
        };
        inputs: Uint8Array;
    };
    loadControlsState(state: any): void;
    saveControlsState(): {
        inputs: Uint8Array;
    };
    getCPUState(): any;
    getRasterScanline(): number;
    getDebugCategories(): string[];
    getDebugInfo(category: any, state: any): string;
}
export {};
